#version 450 core

layout(location=0) out vec4 color;

struct Material {
	sampler2D texture_diffuse1;
	sampler2D texture_specular1;
    sampler2D depth_map1;// the texture corresponding to the height map
	float shininess;
};
uniform Material material;

struct DirLight {
	vec3 direction;
	vec3 ambient;
	vec3 diffuse;
	vec3 specular;
};
uniform mat4 model;
uniform DirLight dirLight;
uniform vec3 CameraPos;
uniform sampler2DShadow shadowMap;
uniform bool smoothShadows = true;

in float height;
in vec3 normal;
in vec2 texCoord;
in vec4 lightSpacePosition;
in vec4 worldPosition;
in mat3 invTransModel;

// Blinn-Phong shading for directional lightsources
vec3 directionalLight(DirLight light, vec3 normal, vec3 viewDir, float directLight, float height) {

	//diffuse
	vec3 lightDir = normalize(-light.direction);
	float diff = max(dot(lightDir, normal), 0.0f);
	vec3 diffuse = light.diffuse * diff;

	//specular
	vec3 halfwayDir = normalize(lightDir + viewDir);  
	float spec = pow(max(dot(normal, halfwayDir), 0.0), 8);
	vec3 specular = light.specular * spec * texture(material.texture_specular1, texCoord*100).r;

	// calculate color based on steepness and height
	float steepness = dot(normal, vec3(0, 1, 0)) * 0.5 + 0.5;
	float a = 10;
	steepness = 0.5 + tanh(a * steepness - a) * 0.5;
	float snowFactor = clamp(height * 20 - 10, 0, 1);
	vec3 grassColor = texture(material.texture_diffuse1, texCoord*100).rgb;
	vec3 stoneColor = vec3(0.2, 0.2, 0.2);
	vec3 snowColor = vec3(1);
	vec3 terrainColor = mix(grassColor, snowColor, snowFactor);
	vec3 texColor = mix(stoneColor, terrainColor, steepness);

	return (light.ambient + directLight * (diffuse + specular)) * texColor;
};

float isFragmentInShadow(vec4 fragPosLightSpace) {
    vec4 projCoords = fragPosLightSpace * 0.5 + 0.5; 

	// compare fragment depths in lightspace
	if (smoothShadows) {
		vec2 texelSize = 1.0 / textureSize(shadowMap, 0);
		float filterKernel[3][3] = {
			{ 0.25, 0.5, 0.25 },
			{ 0.5,  1.0, 0.5 },
			{ 0.25, 0.5, 0.25 }
		};
		float shadow = 0.0;
		for (int i = -1; i <= 1; i++) {
			for (int j = -1; j <= 1; j++) {
				shadow += textureProj(shadowMap, projCoords + vec4(texelSize, 0.0, 0.0) * vec4(i, j, 0, 0));
			}
		}
		return shadow / 4.0;
	} else {
		return textureProj(shadowMap, projCoords);
	}
};

void main() {
    // compute the normal for the fragment
    vec2 delta = 1.0 / textureSize(material.depth_map1, 0);
    float height = texture (material.depth_map1 , texCoord).r;
    float deltaX =  textureOffset(material.depth_map1 , texCoord, ivec2(-1, 0)).r - textureOffset(material.depth_map1 , texCoord, ivec2(1, 0)).r;
    float deltaZ = textureOffset(material.depth_map1 , texCoord, ivec2(0, -1)).r - textureOffset(material.depth_map1 , texCoord, ivec2(0, 1)).r;
    vec3 norm = vec3(deltaX / (2*delta.x), 1, deltaZ / (2*delta.y));
    norm = normalize(invTransModel * norm);
	
	// apply shading
	vec3 viewDir = normalize(CameraPos - worldPosition.xyz);
	float directLight = isFragmentInShadow(lightSpacePosition);
	vec3 lightContribution = directionalLight(dirLight, norm, viewDir, directLight, height);
    color = vec4(lightContribution, 1.0);
}
