#version 450 core

//MATERIAL DES OBJEKTES
struct Material {
	sampler2D texture_diffuse1;
	sampler2D texture_specular1;
	float shininess;
	float emission;
};
uniform Material material;

//POINT LIGHTS
//POSITION IN WORLD SPACE
struct PointLight {
	vec3 position;

	vec3 ambient;
	vec3 diffuse;
	vec3 specular;
	//attenuation
	float constant;
	float linear;
	float quadratic;
};
#define NR_POINT_LIGHTS 4
uniform PointLight pointLights[4];
vec3 CalcPointLight(PointLight light, vec3 normal, vec3 viewDir);


//DIRECTIONAL LIGHTS
//DIRECTION IN WORLD SPACE
struct DirLight {
	vec3 direction;
	vec3 ambient;
	vec3 diffuse;
	vec3 specular;
};
uniform DirLight dirLight;
vec3 CalcDirLight(DirLight light, vec3 normal, vec3 viewDir, float directLight);
float isFragmentInShadow(vec4 fragPosLightSpace, vec3 normal, vec3 lightDir);

uniform sampler2DShadow shadowMap;

//COORDINATES IN WORLD SPACE
//in vec3 FragPos;
//in vec4 FragPosLightSpace;
//in vec2 texCoords;
//in vec3 Normal;

//in VS_FS_INTERFACE {
in vec3 worldPos;
in vec4 lightSpacePos;
in vec2 texCoords;
in vec3 normal;
//} fragment;

uniform vec3 CameraPos;
uniform mat4 view;
uniform bool smoothShadows = true;

out vec4 FragColor;

void main() {

	vec3 finalColor = vec3(0.0f, 0.0f, 0.0f);

	vec3 norm = normalize(normal);

	//viewDir has to be from Fragment to Eye, not Eye to Fragment
	vec3 viewDir = normalize(CameraPos - worldPos);

	// fragment in shadow
	float directLight = isFragmentInShadow(lightSpacePos, norm, dirLight.direction);

	//DIRECTION LIGHT
	finalColor += CalcDirLight(dirLight, norm, viewDir, directLight);

	//POINT LIGHTS
	for(int i = 0; i < NR_POINT_LIGHTS; i++){
		finalColor += CalcPointLight(pointLights[i], norm, viewDir);
	}

	finalColor *= vec3(texture(material.texture_diffuse1, texCoords)); 
	finalColor += material.emission;
	FragColor = vec4(finalColor, 1.0f);
}

// Blinn-Phong shading for directional lightsources
// viewDir has to be from Fragment to Eye, not Eye to Fragment
vec3 CalcDirLight(DirLight light, vec3 normal, vec3 viewDir, float directLight) {
	
	//ambient
	vec3 ambient = light.ambient;

	//diffuse
	vec3 lightDir = normalize(-light.direction);
	vec3 norm = normalize(normal);
	float diff = max(dot(lightDir, norm), 0.0f);
	vec3 diffuse = light.diffuse * diff;

	//specular
	vec3 specular = vec3(0.0);
	if (diff > 0.0) {
		vec3 halfwayDir = normalize(lightDir + viewDir);  
		float spec = pow(max(dot(normal, halfwayDir), 0.0), material.shininess);
		specular = light.specular * spec;
	}

	return ambient + directLight * (diffuse + specular);
};

// Phong shading for point lightsources
vec3 CalcPointLight(PointLight light, vec3 normal, vec3 viewDir) {

	 //attenuation
	float d = length(light.position - worldPos);
	float a = 1.0f / (light.constant + light.linear *d + light.quadratic* pow(d,2));
	
	//ambient
	vec3 ambient = light.ambient;

	//diffuse
	vec3 lightDir = normalize(light.position - worldPos);
	vec3 norm = normalize(normal);
	float diff = max(dot(lightDir, norm), 0.0f);
	vec3 diffuse = light.diffuse * diff;

	//specular
	vec3 reflectDir = normalize(reflect(-lightDir, norm));
	float spec = pow(max(dot(viewDir, reflectDir), 0.0f), material.shininess);
	vec3 specular = light.specular * spec;

	return (ambient + diffuse + specular) *a;
};

// 
float isFragmentInShadow(vec4 fragPosLightSpace, vec3 normal, vec3 lightDir) {

    // perform perspective divide
    vec4 projCoords = fragPosLightSpace * 0.5 + 0.5; 

	// compare fragment depths in lightspace
	if (smoothShadows) {
		vec2 texelSize = 1.0 / textureSize(shadowMap, 0);
		float filterKernel[3][3] = {
			{ 0.25, 0.5, 0.25 },
			{ 0.5,  1.0, 0.5 },
			{ 0.25, 0.5, 0.25 }
		};
		float shadow = 0.0;
		for (int i = -1; i <= 1; i++) {
			for (int j = -1; j <= 1; j++) {
				shadow += textureProj(shadowMap, projCoords + vec4(texelSize, 0.0, 0.0) * vec4(i, j, 0, 0));
			}
		}
		return shadow / 4.0;
	} else {
		return textureProj(shadowMap, projCoords);
	}
};