#version 450 core

in vec2 texCoords;

layout(location=0) out vec4 color;

struct Material {
	sampler2D texture_diffuse1; // actually colormap
	sampler2D texture_specular1; // actually depthmap
};

uniform Material material;
uniform sampler2DShadow shadow_map;
uniform mat4 view_inv;
uniform mat4 projection_inv; // inverse of projection matrix
uniform mat4 light_view;
uniform mat4 light_projection;
uniform vec3 view_pos;
uniform float time;

struct Light {
	int num_samples;
	vec3 position;
	vec3 color;
};
uniform Light light;

float dither_pattern[16] = float[16] ( // “bayer” matrix
	0.0f, 0.5f, 0.125f, 0.625f,
	0.75f, 0.22f, 0.875f, 0.375f,
	0.1875f, 0.6875f, 0.0625f, 0.5625,
	0.9375f, 0.4375f, 0.8125f, 0.3125
);

// noise functions from https://github.com/ashima/webgl-noise

vec3 mod289(vec3 x) {
  return x - floor(x * (1.0 / 289.0)) * 289.0;
}

vec2 mod289(vec2 x) {
  return x - floor(x * (1.0 / 289.0)) * 289.0;
}

vec3 permute(vec3 x) {
  return mod289(((x*34.0)+10.0)*x);
}

float snoise(vec2 v) {
	const vec4 C = vec4(0.211324865405187,  // (3.0-sqrt(3.0))/6.0
						0.366025403784439,  // 0.5*(sqrt(3.0)-1.0)
						-0.577350269189626,  // -1.0 + 2.0 * C.x
						0.024390243902439); // 1.0 / 41.0
	// First corner
	vec2 i  = floor(v + dot(v, C.yy) );
	vec2 x0 = v -   i + dot(i, C.xx);

	// Other corners
	vec2 i1;
	//i1.x = step( x0.y, x0.x ); // x0.x > x0.y ? 1.0 : 0.0
	//i1.y = 1.0 - i1.x;
	i1 = (x0.x > x0.y) ? vec2(1.0, 0.0) : vec2(0.0, 1.0);
	// x0 = x0 - 0.0 + 0.0 * C.xx ;
	// x1 = x0 - i1 + 1.0 * C.xx ;
	// x2 = x0 - 1.0 + 2.0 * C.xx ;
	vec4 x12 = x0.xyxy + C.xxzz;
	x12.xy -= i1;

	// Permutations
	i = mod289(i); // Avoid truncation effects in permutation
	vec3 p = permute( permute( i.y + vec3(0.0, i1.y, 1.0 )) + i.x + vec3(0.0, i1.x, 1.0 ));

	vec3 m = max(0.5 - vec3(dot(x0,x0), dot(x12.xy,x12.xy), dot(x12.zw,x12.zw)), 0.0);
	m = m*m ;
	m = m*m ;

	// Gradients: 41 points uniformly over a line, mapped onto a diamond.
	// The ring size 17*17 = 289 is close to a multiple of 41 (41*7 = 287)
	vec3 x = 2.0 * fract(p * C.www) - 1.0;
	vec3 h = abs(x) - 0.5;
	vec3 ox = floor(x + 0.5);
	vec3 a0 = x - ox;

	// Normalise gradients implicitly by scaling m
	// Approximation of: m *= inversesqrt( a0*a0 + h*h );
	m *= 1.79284291400159 - 0.85373472095314 * ( a0*a0 + h*h );

	// Compute final noise value at P
	vec3 g;
	g.x  = a0.x  * x0.x  + h.x  * x0.y;
	g.yz = a0.yz * x12.xz + h.yz * x12.yw;
	return 130.0 * dot(m, g);
}

float fbm(vec2 xy, int octaves) {
    float result = 0;
    for (int i = 1; i <= octaves; i++) {
        result += snoise(xy * i) / (i);
    }
    return result;
}

// TODO: use uniforms for clipping planes
float linear_depth(float d) {
	float near = 0.1;
	float far = 4000.0;
	return (near * far) / (far + d * (near - far));
}

float get_depth(vec2 uv) {
	return texture(material.texture_specular1, uv).r;
}

vec3 world_pos_from_depth(float depth, vec2 uv) {
	float z = depth * 2.0 - 1.0;
	vec4 clip_space_position = vec4(uv * 2.0 - 1.0, z, 1.0);
	vec4 view_space_position = projection_inv * clip_space_position;
	// Perspective divide
	view_space_position /= view_space_position.w;
	vec4 world_space_position = view_inv * view_space_position;
	return world_space_position.xyz;
}

float sampleShadow(vec4 fragPosLightSpace) {
    vec4 projCoords = fragPosLightSpace * 0.5 + 0.5;
	return textureProj(shadow_map, projCoords);
};

vec3 phase_henyey_greenstein(vec3 g, float cos_theta) {
	const float PI = 3.146;
	vec3 denom = 1 + g * g - 2 * g * cos_theta;
	return 1 / (4 * PI) * (1 - g * g) / (denom * sqrt(denom));
}

vec3 volumetric_light(vec3 frag_pos) {
	vec4 end_pos_worldspace = vec4(view_pos, 1.0); // get ray start and..
	vec4 start_pos_worldspace = vec4(frag_pos, 1.0); // end position in WS (world space)
	vec4 end_pos_lightview = light_view * end_pos_worldspace; // WS -> LS (light space)
	vec4 start_pos_lightview = light_view * start_pos_worldspace; // WS -> LS
	vec4 delta_lightview = normalize(end_pos_lightview - start_pos_lightview); // direction in LS
	vec4 delta_worldspace = normalize(end_pos_worldspace - start_pos_worldspace); // direction in WS
	float raymarch_distance_lightview = length(end_pos_lightview - start_pos_lightview); // s in LS
	float raymarch_distance_worldspace = length(end_pos_worldspace - start_pos_worldspace); // s in W
	float step_size_lightview = raymarch_distance_lightview / light.num_samples; // Δt in LS
	float step_size_worldspace = raymarch_distance_worldspace / light.num_samples; // Δt in WS
	float dither_value = dither_pattern[ (int(gl_FragCoord.x) % 4)*4 +(int(gl_FragCoord.y)%4) ];
	vec4 ray_position_lightview = start_pos_lightview + dither_value * step_size_lightview * delta_lightview;; // current sample position in LS
	vec4 ray_position_worldspace = start_pos_worldspace + dither_value * step_size_worldspace * delta_worldspace;; // current sample position in WS
	vec3 light_contribution = vec3(0);
	float cos_theta = dot(normalize((end_pos_worldspace - start_pos_worldspace).xyz), normalize(-light.position));
	vec3 g = vec3(-0.4, -0.2, -0.0);
	float timeOfDay = (dot(-light.position, vec3(0, 1, 0)) * 0.5 + 0.5);
	//float tau = 0.0001f; // density or propability of collision
	float phi = 50.1f; // light strength
	float PI_RCP = 1 / 3.146;
	for (
		float l = raymarch_distance_worldspace;
		l > step_size_worldspace;
		l -= step_size_worldspace)
	{
		float height = max((ray_position_worldspace.y + 1000), 0); // [m]
		float ground_density = 0.005;
		float density_decay = 0.007; // e.g. haze: 0.8333, molecules: 0.1136 [1/km]
		vec3 add = vec3(0);
		if (height > 1200 && height < 1300) {
			float cloud_noise = clamp(fbm(ray_position_worldspace.xz * vec2(0.7, 1.1) / 2000 + vec2(-0.02 * time), 10) - 0.1, 0, 1);
			add += cloud_noise * phase_henyey_greenstein(g, -cos_theta) * 0.2;
			density_decay += cloud_noise;
		}
		float tau = ground_density * exp(-height * density_decay); // density or propability of collision
		vec4 ray_position_lightclipspace = light_projection * vec4(ray_position_lightview.xyz, 1);
		// check if in shadow or not, and set shadow_term accordingly
		float shadow_term = sampleShadow(ray_position_lightclipspace);
		// get distance in world space
		float d = 1;//length(ray_position_worldspace.xyz - light.position);
		// multiplication is faster than division, so do it once
		float d_rcp = 1.0/d;
		// accumulate to light contribution based on approximation of radiative transport equation
		light_contribution += tau *
			//(shadow_term * (phi * 0.25 * PI_RCP) * d_rcp * d_rcp ) *
			(shadow_term * phi * phase_henyey_greenstein(g, cos_theta)) *
			exp(-d*tau) *
			exp(-l*tau) *
			light.color * 
			timeOfDay * 
			step_size_worldspace
			+ add;
		// continue ray marching
		ray_position_lightview += step_size_lightview * delta_lightview;
		ray_position_worldspace += step_size_worldspace * delta_worldspace;
	}
	return light_contribution;
}

vec3 extinction(float dist) {
	float beta_r = 1.0;
	vec3 beta_m = vec3(0.008, 0.009, 0.01);
	return exp(-beta_m * dist);
}

void main() {
	vec2 uv = texCoords;
	uv.y = uv.y * -1 + 1;
	
	float depth = get_depth(uv);
	vec3 frag_pos = world_pos_from_depth(depth, uv); // construct world position from depth
	float dist = distance(frag_pos, view_pos);

	color = vec4(0, 0, 0, 1);
	vec3 in_scattering = volumetric_light(frag_pos);
	vec3 extinction = extinction(dist);
	color.rgb += vec3(in_scattering);
	//color.rgb += extinction * texture(material.texture_diffuse1, uv).rgb;
}
